import cupy as cp
import numpy as np
import torch
import torch.distributed as dist
from torch.utils.data import Subset, DataLoader
from .kernels import convolute_ctf, highpass2d, project, translate
from .utility import cupy_to_torch, torch_to_cupy

def sieve(dataset, volume, threshold, number, rank, world_size):
    m = len(dataset)
    batch_size = 50

    # Take rank-th part of dataset.
    l, r = round(rank / world_size * m), round((rank + 1) / world_size * m)
    subset = Subset(dataset, list(range(l, r)))
    loader = DataLoader(subset, batch_size, num_workers = 4, pin_memory = True)
    g = torch.zeros(m, dtype = torch.float64, device = 'cuda')
    volume = cp.array(volume, dtype = cp.float64)

    for i_batch, batch in enumerate(loader):

        # Prepare batch data.
        imgs  = torch_to_cupy(batch[0].to('cuda'))
        paras = batch[1].numpy()
        trans = paras[:, 0:2]
        quats = paras[:, 2:6]
        ctfs  = paras[:, 6:14]

        # Compute score.
        imgs = translate(imgs, trans)
        projs = convolute_ctf(project(volume, quats), ctfs) - imgs
        imgs = highpass2d(imgs, threshold)
        projs = highpass2d(projs, threshold)
        lft = l + i_batch * batch_size
        g[lft : lft + len(imgs)] = cupy_to_torch(cp.linalg.norm(projs, axis = (1, 2)) ** 2 - \
                                                 cp.linalg.norm(imgs,  axis = (1, 2)) ** 2)

    if world_size > 1: dist.reduce(g, dst = 0)
    g = g.to('cpu').numpy()
    indices = np.argsort(g)
    mask = np.zeros(m, dtype = np.bool_)
    mask[indices[:number]] = True
    return dataset.subset(mask)



import os
import starfile
import numpy as np
import pandas as pd
from copy import copy
from torch.utils.data import Dataset
from typing import Optional
from .utility import mrcread

class ParticleDataset(Dataset):
    '''
    Dataset class for particles.

    The parameters of particles, like ctfs, will be loaded when
    the object is created. However, the data of particles will not
    be loaded until the __getitem__ method is called.
    '''

    def __init__(self, star_path : str, data_dir : str = '', pixel_size : Optional[float] = None):
        if not os.path.exists(star_path):
            raise FileNotFoundError(f'{star_path} does not exist.')
        star = starfile.read(star_path, always_dict = True)
        self.data_dir = data_dir
        self.pixel_size = pixel_size

        # <Relion 3.1
        # For supporting starfile>=0.5 in the future.
        if len(star) == 1 and (0 in star or '' in star or 'images' in star):
            self.version = 2
            self.optics = None
            self.particles = star[0] if 0 in star else star[''] if '' in star else star['images']

            # Check keys.
            for key in ['rlnOriginX', 'rlnOriginY', 'rlnAngleRot', 'rlnAngleTilt',
                        'rlnAnglePsi', 'rlnVoltage', 'rlnDefocusU', 'rlnDefocusV',
                        'rlnDefocusAngle', 'rlnSphericalAberration',
                        'rlnAmplitudeContrast', 'rlnImageName']:
                if key not in self.particles:
                    raise ValueError(f'Key {key} missed in star file {star_path}.')
            if pixel_size is None:
                raise ValueError('Need pixelsize (--angpix) for star file before RELION version 3.1.')

        # >=Relion 3.1
        elif len(star) == 2 and ('optics' in star and 'particles' in star):
            self.version = 3
            self.optics = star['optics']
            self.particles = star['particles']

            # Check keys.
            for key in ['rlnVoltage', 'rlnImagePixelSize', 'rlnSphericalAberration',
                        'rlnAmplitudeContrast', 'rlnOpticsGroup']:
                if key not in self.optics:
                    raise ValueError(f'Key {key} missed in block data_optics in star file {star_path}.')

            for key in ['rlnOriginXAngst', 'rlnOriginYAngst', 'rlnAngleRot', 'rlnAngleTilt',
                        'rlnAnglePsi', 'rlnDefocusU', 'rlnDefocusV', 'rlnDefocusAngle',
                        'rlnOpticsGroup', 'rlnImageName']:
                if key not in self.particles:
                    raise ValueError(f'Key {key} missed in block data_particles in star file {star_path}.')

        else:
            raise ValueError('Invalid particle star file.')

        self._parse_paras()

    def _parse_paras(self):
        '''
        Parsing parameters from self.optics, self.particles.
        '''
        self.paras = np.empty((len(self.particles), 14), dtype = np.float64)

        # Handling RELION 3.1 format by merging tables.
        if self.version == 2:
            particles = self.particles
        else:
            try:
                particles = pd.merge(self.optics, self.particles, left_on = 'rlnOpticsGroup', right_on = 'rlnOpticsGroup', validate = 'one_to_many')
            except pd.errors.MergeError:
                raise ValueError('There are multiple optic groups with same index. Check the star file.')
            if len(particles) != len(self.particles):
                raise ValueError('There are particles with no corresponding optic group. Check the star file.')

        self.paras[:,  0] = particles['rlnOriginX'] if self.version == 2 else particles['rlnOriginXAngst'] / particles['rlnImagePixelSize']
        self.paras[:,  1] = particles['rlnOriginY'] if self.version == 2 else particles['rlnOriginYAngst'] / particles['rlnImagePixelSize']
        psi   = np.radians(particles['rlnAngleRot'])
        theta = np.radians(particles['rlnAngleTilt'])
        phi   = np.radians(particles['rlnAnglePsi'])
        self.paras[:,  2] =  np.cos((phi + psi) / 2) * np.cos(theta / 2)
        self.paras[:,  3] = -np.sin((phi - psi) / 2) * np.sin(theta / 2)
        self.paras[:,  4] = -np.cos((phi - psi) / 2) * np.sin(theta / 2)
        self.paras[:,  5] = -np.sin((phi + psi) / 2) * np.cos(theta / 2)
        self.paras[:,  6] = particles['rlnVoltage'] * 1e3
        self.paras[:,  7] = particles['rlnDefocusU']
        self.paras[:,  8] = particles['rlnDefocusV']
        self.paras[:,  9] = np.radians(particles['rlnDefocusAngle'])
        self.paras[:, 10] = particles['rlnSphericalAberration'] * 1e7
        self.paras[:, 11] = particles['rlnAmplitudeContrast']
        self.paras[:, 12] = np.radians(particles['rlnPhaseShift']) if 'rlnPhaseShift' in particles else 0.
        self.paras[:, 13] = self.pixel_size if self.version == 2 else particles['rlnImagePixelSize']

    def __len__(self) -> int:
        return len(self.paras)

    def __getitem__(self, i : int):
        assert 0 <= i < len(self.paras)
        slc, name, *_ = self.particles.iloc[i]['rlnImageName'].split('@')
        return mrcread(self.data_dir + name, int(slc) - 1), self.paras[i]

    def save(self, output_path : str):
        data_dict = {'images' : self.particles} if self.version == 2 else \
                    {'optics' : self.optics, 'particles' : self.particles}
        starfile.write(data_dict, output_path, overwrite = True)

    def n_random_subset(self) -> int:
        return 1 if 'rlnRandomSubset' not in self.particles else \
            len(self.particles['rlnRandomSubset'].value_counts())

    def get_random_subset(self, i : int):
        if 'rlnRandomSubset' not in self.particles:
            raise ValueError('Key rlnRandomSubset missed in star file.')
        return self.subset(self.particles['rlnRandomSubset'] == i)

    def subset(self, mask):
        sub = copy(self)
        sub.paras = self.paras[mask]
        sub.particles = self.particles[mask]
        return sub

    def balance(self):
        if 'rlnRandomSubset' not in self.particles:
            return

        n = self.particles['rlnRandomSubset'].value_counts().min()
        self.particles = self.particles.groupby('rlnRandomSubset').sample(n = n)
        self.particles.sort_index(inplace = True)
        self._parse_paras()



import mrcfile
import subprocess
import numpy as np
try:
    import cupy as cp
except:
    pass
from time import time
from torch.utils.dlpack import to_dlpack, from_dlpack

def check_cupy():
    try:
        import cupy
    except ModuleNotFoundError:
        print('[ERROR] CuPy module not found.')
        exit(1)
    except ImportError:
        print('[ERROR] Error occured when importing CuPy. Please check your CUDA environment or GPU card.')
        exit(1)

def cupy_to_torch(x):
    return from_dlpack(x.toDlpack())

def torch_to_cupy(x):
    return cp.fromDlpack(to_dlpack(x))

def mrcread(fpath : str, iSlc = None):
    with mrcfile.mmap(fpath, permissive = True, mode = 'r') as mrc:
        data = mrc.data if iSlc is None or mrc.data.ndim == 2 else mrc.data[iSlc]
        return np.array(data, dtype = np.float64)

def run_commands(commands, msg = '', stdout = None):
    time0 = time()
    if isinstance(commands, str): commands = [commands]
    processes = [subprocess.Popen(command, shell = True, stdout = stdout) for command in commands]
    for process in processes: process.wait()
    time1 = time()
    if all(process.returncode == 0 for process in processes):
        print(f'[Subprocesses completed successfully in {time1 - time0:.2f}s.][{msg}]')
    else:
        print(f'[Subprocesses reported an error.',
              f'Return code: {[process.returncode for process in processes]}.][{msg}]')
        exit()



import argparse
import csv
import os
import re
import subprocess
import sys
from itertools import product
from pathlib import Path
from concurrent.futures import ThreadPoolExecutor
from threading import Lock
from time import sleep
from random import uniform

def parse_argument():
    parser = argparse.ArgumentParser(description = 'cryosieve-csrefine: automatic SPA 3D-refinement by calling CryoSPARC.')
    parser.add_argument('--i',         type = str, nargs    = '+',  help = 'input star file(s) or txt file(s) containing a list of star files.')
    parser.add_argument('--directory', type = str, default  = '',   help = 'directory of particles, empty (current directory) by default.')
    parser.add_argument('--o',         type = str,                  help = 'output summary csv file path. If not provided, no summary is written.')
    parser.add_argument('--sym',       type = str, default  = 'C1', help = 'molecular symmetry, C1 by default.')
    parser.add_argument('--ref',       type = str,                  help = 'initial reference model. If not provided, CryoSPARC\'s ab-initio job will be used.')
    parser.add_argument('--ini_high',  type = float,                help = 'initial resolution.')
    parser.add_argument('--repeat',    type = int, default  = 1,    help = 'number of trials, 1 by default.')
    parser.add_argument('--user',      type = str, required = True, help = 'e-mail address of the user of CryoSPARC.')
    parser.add_argument('--project',   type = str, required = True, help = 'project UID in CryoSPARC.')
    parser.add_argument('--workspace', type = str, required = True, help = 'workspace UID in CryoSPARC.')
    parser.add_argument('--lane',      type = str, required = True, help = 'lane selected for computing in CryoSPARC.')
    parser.add_argument('--nu',        action = 'store_true',       help = 'use non-uniform refinement.')
    parser.add_argument('--local',     action = 'store_true',       help = 'use local refinement after homogeneous / non-uniform refinement.')
    parser.add_argument('--min_angular_step', type = float, default = 0.01, help = 'minimum angular step for local refinement.')
    parser.add_argument('--resplit',   action = 'store_true',       help = 'force re-do GS split.')
    parser.add_argument('--workers',   type = int,                  help = 'number of workers to run CryoSPARC job, unlimited by default.')
    if len(sys.argv) == 1:
        parser.print_help()
        exit()
    return parser.parse_args()

def load_cryosparc(args):
    from cryosparc_compute.client import CommandClient
    global host, port, client, user_id, project_uid, workspace_uid, lane, lock

    host = os.environ['CRYOSPARC_MASTER_HOSTNAME']
    port = os.environ['CRYOSPARC_COMMAND_CORE_PORT']
    client = CommandClient(host = host, port = port)
    user_id = client.get_id_by_email(args.user)
    project_uid = args.project
    workspace_uid = args.workspace
    lane = args.lane
    lock = Lock()

def enqueue_and_wait(local_lane, **kwargs):
    with lock:
        job_id = client.make_job(
            user_id = user_id,
            project_uid = project_uid,
            workspace_uid = workspace_uid,
            **kwargs
        )
        client.enqueue_job(project_uid, job_id, local_lane, user_id)

    while True:
        sleep(uniform(20, 40))
        with lock: job_status = client.get_job_status(project_uid, job_id)
        if job_status == 'completed':
            return job_id
        elif job_status in ['failed', 'killed']:
            raise RuntimeError(f'Job {job_id} is {job_status}.')

def import_particles(particle_meta_path):
    import_job_id = enqueue_and_wait(
        None,
        job_type = 'import_particles',
        params = {
            'particle_meta_path' : particle_meta_path,
            'particle_blob_path' : str(Path(args.directory).absolute())
        }
    )
    return import_job_id

def abinit(import_particles_job_id):
    abinit_job_id = enqueue_and_wait(
        lane,
        job_type = 'homo_abinit',
        input_group_connects = {
            'particles': f'{import_particles_job_id}.imported_particles'
        }
    )
    return abinit_job_id

def refine(particle, volume):
    params = {
        'refine_symmetry' : args.sym,
        'refine_gs_resplit' : args.resplit
    }
    if args.ini_high is not None: params['refine_res_init'] = args.ini_high
    refine_job_id = enqueue_and_wait(
        lane,
        job_type = 'nonuniform_refine_new' if args.nu else 'homo_refine_new',
        params = params,
        input_group_connects = {
            'particles' : particle,
            'volume' : volume
        }
    )
    return refine_job_id

def local_refine(refine_job_id):
    local_refine_job_id = enqueue_and_wait(
        lane,
        job_type = 'new_local_refine',
        params = {
            'refine_symmetry' : args.sym,
            'use_alignment_prior' : True,
            'min_angular_step' : args.min_angular_step
        },
        input_group_connects = {
            'particles' : f'{refine_job_id}.particles',
            'volume' : f'{refine_job_id}.volume',
            'mask' : f'{refine_job_id}.mask'
        }
    )
    return local_refine_job_id

def parse_result(refine_job_id):
    with lock: streamlog = client.get_job_streamlog(project_uid, refine_job_id)
    text_list = [entry['text'].strip() for entry in streamlog if 'text' in entry]
    resolution = bfactor = None
    for text in text_list:
        if text.startswith('Using Filter Radius'):
            resolution = re.search(r'\(([\d.]+)A\)', text).group(1)
        elif text.startswith('Estimated Bfactor:'):
            bfactor = re.search(r'(-[\d.]+)', text).group(1)
        elif text.startswith('Split A has'):
            num_A = int(re.search(r'\b(\d+)\b', text).group(1))
        elif text.startswith('Split B has'):
            num_B = int(re.search(r'\b(\d+)\b', text).group(1))

    with lock: alignment3D_path = client.get_project_dir_abs(project_uid) + '/' + client.get_job_result(project_uid, f'{refine_job_id}.particles.alignments3D')['metafile']
    return [num_A + num_B, resolution, bfactor, alignment3D_path]

def check_results(results, info = None):
    if info is not None:
        print(info, results)
    if any(result is None for result in results):
        raise RuntimeError('Error occurred during execution.')

def main():
    args = parse_argument()
    input_path = ' '.join(f'"{file}"' for file in args.i)
    command = f'eval $(cryosparcm env) && python {Path(__file__).absolute()} ' + \
              f'--i {input_path} ' + \
              (f'--directory "{args.directory}" ' if args.directory else '') + \
              (f'--o "{args.o}" ' if args.o is not None else '') + \
              f'--sym {args.sym} ' + \
              (f'--ref "{args.ref}" ' if args.ref is not None else '') + \
              (f'--ini_high {args.ini_high} ' if args.ini_high is not None else '') + \
              f'--repeat {args.repeat} ' + \
              f'--user {args.user} ' + \
              f'--project {args.project} ' + \
              f'--workspace {args.workspace} ' + \
              f'--lane {args.lane} ' + \
              ('--nu ' if args.nu else '') + \
              ('--local ' if args.local else '') + \
              ('--resplit ' if args.resplit else '') + \
              (f'--workers {args.workers} ' if args.workers is not None else '')
    subprocess.run(command, shell = True)

def parse_meta_paths(paths):
    particle_meta_paths = []
    for path in paths:
        path = Path(path)
        if path.suffix == '.star':
            particle_meta_paths.append(str(path.absolute()))
        elif path.suffix == '.txt':
            with open(path) as fin: lines = fin.readlines()
            for line in lines:
                meta_path = Path(line.strip())
                if meta_path.suffix != '.star':
                    raise RuntimeError(f'{meta_path} is not a star file.')
                particle_meta_paths.append(str(meta_path.absolute()))
        else:
            raise RuntimeError(f'"{path}" is not a star or txt file.')
    return particle_meta_paths

if __name__ == '__main__':
    args = parse_argument()
    load_cryosparc(args)

    # Check args.
    particle_meta_paths = parse_meta_paths(args.i)
    if not Path(args.directory).is_dir():
        raise FileNotFoundError(f'"{args.directory}" is not a directory.')
    if args.ref is not None:
        if not Path(args.ref).is_file():
            raise FileNotFoundError(f'"{args.ref}" does not exist.')

    # Setup.
    pool = ThreadPoolExecutor() if args.workers is None else ThreadPoolExecutor(max_workers = args.workers)

    # Import particles.
    import_particles_tasks = [pool.submit(import_particles, particle_meta_path) for particle_meta_path in particle_meta_paths]
    import_particles_job_ids = [task.result() for task in import_particles_tasks]
    check_results(import_particles_job_ids, 'Import particles:')
    particles = [f'{import_particles_job_id}.imported_particles' for import_particles_job_id in import_particles_job_ids]

    # Import volumes, if ref is given.
    if args.ref is not None:
        import_ref_job_id = enqueue_and_wait(
            None,
            job_type = 'import_volumes',
            params = {
                'volume_blob_path' : str(Path(args.ref).absolute())
            }
        )
        check_results([import_ref_job_id], 'Imported reference maps:')
        volumes = [f'{import_ref_job_id}.imported_volume_1'] * len(particles)
    # Otherwise, use ab-initio reconstruction.
    else:
        abinit_tasks = [pool.submit(abinit, import_particles_job_id) for import_particles_job_id in import_particles_job_ids]
        abinit_job_ids = [task.result() for task in abinit_tasks]
        check_results(abinit_job_ids, 'Ab-initio reconstruction:')
        volumes = [f'{abinit_job_id}.volume_class_0' for abinit_job_id in abinit_job_ids]

    # Homogenous / non-uniform refinement.
    refine_tasks = [pool.submit(refine, particle, volume) for (particle, volume), _ in product(zip(particles, volumes), range(args.repeat))]
    refine_job_ids = [task.result() for task in refine_tasks]
    check_results(refine_job_ids, 'Homogeneous / Non-uniform refinement:')

    # Local refinement.
    if args.local:
        refine_tasks = [pool.submit(local_refine, refine_job_id) for refine_job_id in refine_job_ids]
        refine_job_ids = [task.result() for task in refine_tasks]
        check_results(refine_job_ids, 'Local refinement:')

    # Get results.
    parse_result_tasks = [pool.submit(parse_result, refine_job_id) for refine_job_id in refine_job_ids]
    results = [task.result() for task in parse_result_tasks]
    check_results(results)

    # Shutdown.
    pool.shutdown()

    # Save results.
    with open(args.o, 'w') as csvfile:
        csvwriter = csv.writer(csvfile)
        csvwriter.writerow(['filename', 'number of particles', 'resolution (A)', 'b-factor (A^2)', 'alignment3D'])
        for i, result in enumerate(results):
            particle_meta_path = particle_meta_paths[i // args.repeat]
            csvwriter.writerow([particle_meta_path] + result)



import argparse
import sys

def parse_argument():
    parser = argparse.ArgumentParser(description = 'CryoSieve: a particle sorting and sieving software for single particle analysis in cryo-EM.')
    parser.add_argument('--reconstruct_software', type = str,   required = True,  help = 'command for reconstruction.')
    parser.add_argument('--postprocess_software', type = str,   required = False, help = 'command for postprocessing.')
    parser.add_argument('--i',                    type = str,   required = True,  help = 'input star file path.')
    parser.add_argument('--o',                    type = str,   required = True,  help = 'output path prefix.')
    parser.add_argument('--angpix',               type = float, required = True,  help = 'pixelsize in Angstrom.')
    parser.add_argument('--sym',                  type = str,   default  = 'C1',  help = 'molecular symmetry, C1 by default.')
    parser.add_argument('--num_iters',            type = int,   default  = 10,    help = 'number of iterations for applying CryoSieve, 10 by default.')
    parser.add_argument('--frequency_start',      type = float, default  = 50.,   help = 'starting threshold frquency, in Angstrom, 50A by default.')
    parser.add_argument('--frequency_end',        type = float, default  = 3.,    help = 'ending threshold frquency, in Angstrom, 3A by default.')
    parser.add_argument('--retention_ratio',      type = float, default  = 0.8,   help = 'fraction of retained particles in each iteration, 0.8 by default.')
    parser.add_argument('--mask',                 type = str,   required = True,  help = 'mask file path.')
    parser.add_argument('--balance',              action = 'store_true',          help = 'randomly drop particles to make all subset into the same size.')
    parser.add_argument('--num_gpus',             type = int,   default  = 1,     help = 'number of gpus to execute CryoSieve core program, 1 by default.')
    if len(sys.argv) == 1:
        parser.print_help()
        exit()
    return parser.parse_args()

def main():
    args = parse_argument()
    if args.postprocess_software is not None:
        print('The `--postprocess_software` will be deprecated soon. CryoSieve will implement its own postprocessing.', file = sys.stderr)

    from .utility import check_cupy
    check_cupy()

    import numpy as np
    from pathlib import Path
    from .ParticleDataset import ParticleDataset
    from .utility import run_commands

    # prepare.
    src = Path(f'{args.i}')
    dst = Path(f'{args.o}_iter0.star')
    if not src.exists():
        raise FileNotFoundError(f'{args.i} not found.')
    elif src.suffix != '.star':
        raise ValueError(f'{args.i} is not a star file.')
    dst.absolute().parent.mkdir(parents = True, exist_ok = True)

    dataset = ParticleDataset(src, '', args.angpix)
    if args.balance: dataset.balance()
    dataset.save(dst)

    # go.
    frequences = 1 / np.linspace(1.0 / args.frequency_start, 1.0 / args.frequency_end, args.num_iters)
    overall_retention_ratio = 1.0
    for i in range(args.num_iters):
        print(f'[ITER {i}][Overall Retaining ratio: {overall_retention_ratio * 100:.2f}%][Threshold frequency : {frequences[i]:.2f}A]')

        # reconstruct.
        commands = [f'{args.reconstruct_software} --i {args.o}_iter{i}.star --o {args.o}_iter{i}_half1.mrc --angpix {args.angpix} --sym {args.sym} --ctf true --subset 1 >{args.o}_iter{i}_reconstruct_half1.txt',
                    f'{args.reconstruct_software} --i {args.o}_iter{i}.star --o {args.o}_iter{i}_half2.mrc --angpix {args.angpix} --sym {args.sym} --ctf true --subset 2 >{args.o}_iter{i}_reconstruct_half2.txt']
        run_commands(commands, f'RECONSTRUCT_DONE, ITERATION {i}')

        # postprocess.
        if args.postprocess_software is not None:
            Path(f'{args.o}_postprocess_iter{i}').mkdir(parents = True, exist_ok = True)
            run_commands(f'{args.postprocess_software} --mask {args.mask} --i {args.o}_iter{i}_half1.mrc --i2 {args.o}_iter{i}_half2.mrc --o {args.o}_postprocess_iter{i}/iter{i} --angpix {args.angpix} --auto_bfac --autob_lowres 10 >{args.o}_postprocess_iter{i}.txt', f'POSTPROCESS_DONE, ITERATION {i}')

        # sieve.
        run_commands(f'cryosieve-core --i {args.o}_iter{i}.star --o {args.o}_iter{i + 1}.star --angpix {args.angpix} --volume {args.o}_iter{i}_half1.mrc --volume {args.o}_iter{i}_half2.mrc --mask {args.mask} --retention_ratio {args.retention_ratio} --frequency {frequences[i]:.3f} --num_gpus {args.num_gpus} >{args.o}_iter{i}_sieve.txt', f'SIEVE_DONE, ITERATION {i}')
        overall_retention_ratio *= args.retention_ratio

    print('EXECUTION IN SUCCESS')

if __name__ == '__main__':
    main()



__all__ = [
    'convolute_ctf',
    'highpass2d',
    'project',
    'translate'
]

import cupy as cp
from cupy.fft import rfftn, irfftn

BLOCKSIZE = 1024
BLOCKDIM = lambda x : (x - 1) // BLOCKSIZE + 1

def convolute_ctf(stack, ctfs, order = 1):
    ker_convolute_ctf = cp.RawKernel(r'''
extern "C" __global__ void convolute_ctf(
    double* data,
    int m,
    int n,
    const double* ctfs,
    int order)
{
    int tid = blockDim.x * blockIdx.x + threadIdx.x;
    int n_ = n / 2 + 1;
    if (tid < m * n * n_) {
        int x = tid % n_;
        int y = tid / n_ % n; y = y < n / 2 ? y : y - n;
        int z = tid / n_ / n;

        double voltage           = ctfs[z * 8    ];
        double defocusU          = ctfs[z * 8 + 1];
        double defocusV          = ctfs[z * 8 + 2];
        double astigmatism       = ctfs[z * 8 + 3];
        double Cs                = ctfs[z * 8 + 4];
        double amplitudeContrast = ctfs[z * 8 + 5];
        double phaseShift        = ctfs[z * 8 + 6];
        double pixelSize         = ctfs[z * 8 + 7];

        double pi = 3.1415926535897932384626;
        double waveLength = 12.2643247 / sqrt(voltage * (1 + voltage * 0.978466e-6));
        double f = hypot(x / (pixelSize * n), y / (pixelSize * n));
        double alpha = atan2((double)y, (double)x) - astigmatism;
        double defocus = -(defocusU + defocusV + (defocusU - defocusV) * cos(2 * alpha)) / 2;
        double chi = pi * waveLength * defocus * pow(f, 2.) + pi / 2 * Cs * pow(waveLength, 3.) * pow(f, 4.) - phaseShift;
        double u = pow(-sqrt(1 - pow(amplitudeContrast, 2.)) * sin(chi) + amplitudeContrast * cos(chi), (double)order);

        data[tid * 2    ] *= u;
        data[tid * 2 + 1] *= u;
    }
}''', 'convolute_ctf')

    assert isinstance(stack, cp.ndarray) and stack.ndim == 3 and stack.shape[1] == stack.shape[2]
    m, n, _ = stack.shape
    ctfs = cp.array(ctfs, dtype = cp.float64)
    assert ctfs.shape == (m, 8)

    f_stack = rfftn(stack, axes = (1, 2))
    ker_convolute_ctf((BLOCKDIM(f_stack.size), ), (BLOCKSIZE, ), (f_stack, m, n, ctfs, order))
    return irfftn(f_stack, axes = (1, 2))

def highpass2d(stack, threshold):
    ker_highpass2d = cp.RawKernel(r'''
extern "C" __global__ void highpass2d(
    double* data,
    int m,
    int n,
    double threshold)
{
    int tid = blockDim.x * blockIdx.x + threadIdx.x;
    int n_ = n / 2 + 1;
    if (tid < m * n * n_) {
        int x = tid % n_;
        int y = tid / n_ % n; y = y < n / 2 ? y : y - n;

        double f = hypot((double)x / n, (double)y / n);
        if (f < threshold) {
            data[tid * 2    ] = 0;
            data[tid * 2 + 1] = 0;
        }
    }
}''', 'highpass2d')

    assert isinstance(stack, cp.ndarray) and stack.ndim == 3 and stack.shape[1] == stack.shape[2]
    m, n, _ = stack.shape

    f_stack = rfftn(stack, axes = (1, 2))
    ker_highpass2d((BLOCKDIM(f_stack.size), ), (BLOCKSIZE, ), (f_stack, m, n, threshold))
    return irfftn(f_stack, axes = (1, 2))

def project(volume, quats):
    ker_project = cp.RawKernel(r'''
extern "C" __global__ void project(
    const double* vol,
    int n,
    double qw,
    double qx,
    double qy,
    double qz,
    double* img)
{
    int tid = blockDim.x * blockIdx.x + threadIdx.x;
    if (tid < n * n * n) {
        int x = tid % n     - n / 2;
        int y = tid / n % n - n / 2;
        int z = tid / n / n - n / 2;

        double vx = x - 2 * (qy * qy + qz * qz) * x + 2 * (qx * qy - qw * qz) * y + 2 * (qx * qz + qw * qy) * z + n / 2;
        double vy = y + 2 * (qx * qy + qw * qz) * x - 2 * (qx * qx + qz * qz) * y + 2 * (qy * qz - qw * qx) * z + n / 2;
        // double vz = z + 2 * (qx * qz - qw * qy) * x + 2 * (qy * qz + qw * qx) * y - 2 * (qx * qx + qy * qy) * z + n / 2;
        x = floor(vx);
        y = floor(vy);
        double dx = vx - x;
        double dy = vy - y;

        double voxel = vol[tid];
        if (0 <= x     && x     < n && 0 <= y     && y     < n) atomicAdd(img + (y    ) * n + (x    ), voxel * (1 - dx) * (1 - dy));
        if (0 <= x + 1 && x + 1 < n && 0 <= y     && y     < n) atomicAdd(img + (y    ) * n + (x + 1), voxel * (    dx) * (1 - dy));
        if (0 <= x     && x     < n && 0 <= y + 1 && y + 1 < n) atomicAdd(img + (y + 1) * n + (x    ), voxel * (1 - dx) * (    dy));
        if (0 <= x + 1 && x + 1 < n && 0 <= y + 1 && y + 1 < n) atomicAdd(img + (y + 1) * n + (x + 1), voxel * (    dx) * (    dy));
    }
}''', 'project')

    assert isinstance(volume, cp.ndarray) and volume.ndim == 3
    m, n = len(quats), volume.shape[0]
    assert volume.shape == (n, n, n)

    stack = cp.zeros((m, n, n), dtype = cp.float64)
    for i in range(m):
        qw, qx, qy, qz = quats[i]
        ker_project((BLOCKDIM(volume.size), ), (BLOCKSIZE, ), (volume, n, qw, qx, qy, qz, stack[i]))
    return stack

def translate(stack, trans):
    ker_translate = cp.RawKernel(r'''
extern "C" __global__ void translate(
    double* data,
    int m,
    int n,
    const double* trans)
{
    int tid = blockDim.x * blockIdx.x + threadIdx.x;
    int n_ = n / 2 + 1;
    if (tid < m * n * n_) {
        int x = tid % n_;
        int y = tid / n_ % n; y = y < n / 2 ? y : y - n;
        int z = tid / n_ / n;

        double pi = 3.1415926535897932384626;
        double ax = data[2 * tid    ];
        double ay = data[2 * tid + 1];
        double tx = trans[2 * z    ];
        double ty = trans[2 * z + 1];
        double phi = -2 * pi * (tx * x / n + ty * y / n);
        double cosphi = cos(phi);
        double sinphi = sin(phi);
        data[2 * tid    ] = ax * cosphi - ay * sinphi;
        data[2 * tid + 1] = ax * sinphi + ay * cosphi;
    }
}''', 'translate')

    assert isinstance(stack, cp.ndarray) and stack.ndim == 3 and stack.shape[1] == stack.shape[2]
    m, n, _ = stack.shape
    trans = cp.array(trans, dtype = cp.float64)
    assert trans.shape == (m, 2)

    f_stack = rfftn(stack, axes = (1, 2))
    ker_translate((BLOCKDIM(f_stack.size), ), (BLOCKSIZE, ), (f_stack, m, n, trans))
    return irfftn(f_stack, axes = (1, 2))



import argparse
import os
import sys
from time import time

def core_parser():
    parser = argparse.ArgumentParser(description = 'CryoSieve core.')
    parser.add_argument('--i',               type = str,   required = True, help = 'input star file path.')
    parser.add_argument('--o',               type = str,   required = True, help = 'output star file path.')
    parser.add_argument('--directory',       type = str,   default  = '',   help = 'directory of particles, empty (current directory) by default.')
    parser.add_argument('--angpix',          type = float,                  help = 'pixelsize in Angstrom.')
    parser.add_argument('--volume',          type = str,   required = True, action = 'append', help = 'list of volume file paths.')
    parser.add_argument('--mask',            type = str,                    help = 'mask file path.')
    parser.add_argument('--retention_ratio', type = float, required = True, help = 'fraction of retained particles, 0.8 by default.')
    parser.add_argument('--frequency',       type = float, required = True, help = 'cut-off highpass frequency.')
    return parser

def main(args):
    import torch
    import torch.distributed
    import numpy as np
    import cupy as cp
    from .ParticleDataset import ParticleDataset
    from .utility import mrcread
    from .sieve import sieve

    # Initialize.
    is_distributed = 'RANK' in os.environ and 'WORLD_SIZE' in os.environ
    rank       = int(os.environ['RANK']) if is_distributed else 0
    world_size = int(os.environ['WORLD_SIZE']) if is_distributed else 1
    if is_distributed:
        if torch.distributed.is_nccl_available():
            torch.distributed.init_process_group('nccl')
        elif torch.distributed.is_gloo_available():
            torch.distributed.init_process_group('gloo')
        else:
            raise RuntimeError('Need nccl or gloo for multi-GPU cryosieve-core.')
    torch.cuda.set_device(rank)
    cp.cuda.runtime.setDevice(rank)

    # Input.
    dataset     = ParticleDataset(args.i, args.directory, args.angpix)
    volumes     = [mrcread(path) for path in args.volume]
    mask        = mrcread(args.mask)
    volumes     = [volume * mask for volume in volumes]
    ratio       = args.retention_ratio
    threshold   = args.angpix / args.frequency
    output_path = args.o

    # Process.
    n_subset = dataset.n_random_subset()
    if n_subset != len(volumes):
        raise ValueError('Number of particle subsets should be the same as number of input volumes.')

    mask = np.zeros(len(dataset), dtype = np.bool_)
    for i in range(n_subset):
        subset = dataset.get_random_subset(i + 1)
        if rank == 0:
            print(f'Processing subset {i}.')
            print(f'There are {len(subset)} particles.')
        n_rem = round(ratio * len(subset))
        subset_rem = sieve(subset, volumes[i], threshold, n_rem, rank, world_size)
        if rank == 0:
            mask[subset_rem.particles.index] = True
            print(f'{n_rem} particles remained.\n')

    if rank == 0:
        dataset_rem = dataset.subset(mask)
        dataset_sie = dataset.subset(~mask)
        root, ext = os.path.splitext(output_path)
        dataset_rem.save(output_path)
        dataset_sie.save(root + '_sieved' + ext)

def core():
    parser = core_parser()
    parser.add_argument('--num_gpus',        type = int,   default  = 1,     help = 'number of GPUs to execute the cryosieve program, 1 by default.')
    if len(sys.argv) == 1:
        parser.print_help()
        exit()
    args = parser.parse_args()

    from .utility import check_cupy
    check_cupy()

    from .utility import run_commands
    if args.num_gpus == 1:
        time0 = time()
        main(args)
        time1 = time()
        print(f'[Processes completed successfully in {time1 - time0:.2f}s.][SINGLE-GPU CRYOSIEVE-CORE DONE.]')
    else:
        core_args = f'--i {args.i} --o {args.o} {f"--directory {args.directory}" if args.directory else ""} --angpix {args.angpix} {" ".join([f"--volume {volume}" for volume in args.volume])} {f"--mask {args.mask}" if args.mask is not None else ""} --retention_ratio {args.retention_ratio} --frequency {args.frequency}'
        run_commands(f'torchrun --standalone --nnodes=1 --nproc_per_node={args.num_gpus} -m cryosieve.core ' + core_args, 'MULTI-GPU CRYOSIEVE-CORE DONE.')

if __name__ == '__main__':
    main(core_parser().parse_args())



import argparse
import csv
import starfile
import subprocess
import sys
import tempfile
import numpy as np
import pandas as pd
from scipy.optimize import curve_fit
from .cs_refine import parse_meta_paths

def parse_argument():
    parser = argparse.ArgumentParser(description = 'cryosieve-csrhbfactor: automatic Rosenthal-Henderson B-factor estimation by calling CryoSPARC.')
    parser.add_argument('--i',         type = str, nargs    = '+',  help = 'input star file(s) or txt file(s) containing a list of star files.')
    parser.add_argument('--directory', type = str, default  = '',   help = 'directory of particles, empty (current directory) by default.')
    parser.add_argument('--o',         type = str, required = True, help = 'output summary csv file path.')
    parser.add_argument('--sym',       type = str, default  = 'C1', help = 'molecular symmetry, C1 by default.')
    parser.add_argument('--ref',       type = str,                  help = 'initial reference model. If not provided, CryoSPARC\'s ab-initio job will be used.')
    parser.add_argument('--ini_high',  type = float,                help = 'initial resolution.')
    parser.add_argument('--voltage',   type = int, default  = 300,  help = 'acceleration voltage (kV), 300 by default. Only 200 and 300 supported!')
    parser.add_argument('--repeat',    type = int, default  = 1,    help = 'number of trials, 1 by default.')
    parser.add_argument('--halves',    type = int, default  = 4,    help = 'number of times executing halvings, 4 by default.')
    parser.add_argument('--user',      type = str, required = True, help = 'e-mail address of the user of CryoSPARC.')
    parser.add_argument('--project',   type = str, required = True, help = 'project UID in CryoSPARC.')
    parser.add_argument('--workspace', type = str, required = True, help = 'workspace UID in CryoSPARC.')
    parser.add_argument('--lane',      type = str, required = True, help = 'lane selected for computing in CryoSPARC.')
    parser.add_argument('--nu',        action = 'store_true',       help = 'use non-uniform refinement.')
    parser.add_argument('--resplit',   action = 'store_true',       help = 'force re-do GS split.')
    parser.add_argument('--workers',   type = int,                  help = 'number of workers to run CryoSPARC job, unlimited by default.')
    if len(sys.argv) == 1:
        parser.print_help()
        exit()
    return parser.parse_args()

def main():
    args = parse_argument()

    # Check args.
    particle_meta_paths = parse_meta_paths(args.i)
    if args.voltage not in [200, 300]:
        raise ValueError('Only 200 or 300kV acceleration voltage supported.')
    if len(args.sym) == 0 or args.sym[0] not in ['C', 'D', 'T', 'O', 'I']:
        raise ValueError('Invalid molecular symmetry.')
    if args.sym[0] == 'C':
        N = int(args.sym[1:])
    elif args.sym[0] == 'D':
        N = 2 * int(args.sym[1:])
    elif args.sym[0] == 'T':
        N = 12
    elif args.sym[0] == 'O':
        N = 24
    elif args.sym[0] == 'I':
        N = 60
    else:
        raise ValueError('Invalid molecular symmetry.')
    csvfile = open(args.o, 'w')
    csvwriter = csv.writer(csvfile)

    # For each star file, take random halves and save it to a temporary file.
    tmp_dir = tempfile.TemporaryDirectory()
    with open(f'{tmp_dir.name}/list.txt', 'w') as lst_file:
        for i, meta_path in enumerate(particle_meta_paths):
            star = starfile.read(meta_path, always_dict = True)
            if len(star) == 1 and (0 in star or '' in star or 'images' in star):
                optics = None
                particles = star[0] if 0 in star else star[''] if '' in star else star['images']
            elif len(star) == 2 and ('optics' in star and 'particles' in star):
                optics = star['optics']
                particles = star['particles']
            else:
                raise ValueError('Invalid particle star file.')

            for j in range(args.repeat):
                for k in range(args.halves + 1):
                    particles_j = particles.sample(frac = 0.5 ** k) if args.resplit else \
                                  pd.concat([particles[particles['rlnRandomSubset'] == 1].sample(frac = 0.5 ** k),
                                             particles[particles['rlnRandomSubset'] == 2].sample(frac = 0.5 ** k)])
                    tmp_file = f'{tmp_dir.name}/{i}_{j}_{k}.star'
                    if optics is None:
                        starfile.write({'images' : particles_j}, tmp_file, overwrite = True)
                    else:
                        starfile.write({'optics' : optics, 'particles' : particles_j}, tmp_file, overwrite = True)
                    print(tmp_file, file = lst_file)

    # Call cryosieve-csrefine to estimate resolutions.
    command = f'cryosieve-csrefine --i {tmp_dir.name}/list.txt ' + \
              (f'--directory "{args.directory}" ' if args.directory else '') + \
              f'--o {tmp_dir.name}/summary.csv ' + \
              f'--sym {args.sym} ' + \
              (f'--ref "{args.ref}" ' if args.ref is not None else '') + \
              (f'--ini_high {args.ini_high} ' if args.ini_high is not None else '') + \
              f'--user {args.user} ' + \
              f'--project {args.project} ' + \
              f'--workspace {args.workspace} ' + \
              f'--lane {args.lane} ' + \
              ('--nu ' if args.nu else '') + \
              ('--resplit ' if args.resplit else '') + \
              (f'--workers {args.workers} ' if args.workers is not None else '')
    process = subprocess.run(command, shell = True)
    if process.returncode:
        raise RuntimeError('Error in running cryosieve-csrefine.')

    # Fit RH-curve.
    data = pd.read_csv(f'{tmp_dir.name}/summary.csv')
    num_points = args.repeat * (args.halves + 1)
    rh_bfactor_curve = lambda x, a : np.log(1437.5695 if args.voltage else 1831.7256) + a / (2 * x ** 2) - np.log(x)

    # Save results.
    csvwriter.writerow(['filename', 'RH-b-factor (A^2)'])
    for i, meta_path in enumerate(particle_meta_paths):
        i_slice = slice(i * num_points, (i + 1) * num_points)
        lognum_particles = np.log(N * data.iloc[i_slice]['number of particles'])
        resolutions = data.iloc[i_slice]['resolution (A)']
        rh_bfactor = -curve_fit(rh_bfactor_curve, resolutions, lognum_particles)[0][0]
        csvwriter.writerow([meta_path, rh_bfactor])
    csvfile.close()
